/** 
 * \file proposal.hpp  
 * This file is part of fmcs.
 * Copyright David Rushing Dewhurst, 2022 - present.
 * Some rights reserved.
 */

#ifndef PROPOSAL_H
#define PROPOSAL_H

#include <iostream>

#include <utility>
#include <variant>

#include <functional_util.hpp>
#include <record.hpp>
#include <query.hpp>

/**
 * @brief Type of values returned from filtering algorithms
 * 
 * @tparam O The output type of a probabilistic program
 * @tparam Ts The types of the distributions in the probabilistic program
 */
template<typename O, typename... Ts>
using FilterValueType = std::shared_ptr<record_collection_t<O, Ts...>>;

/**
 * @brief Creates a bare node of type node_t<Out> from a record of nearly arbitrary type.
 * 
 * The node in the input record at the specified address must have a value type equal to 
 * the value type generated by Out.
 * 
 * @tparam Out the underlying distribution in the output node
 * @tparam In the collection of distributions by which the input record is parameterized
 * @param r_in the input record
 * @param address the address in the input record to access
 * @return node_t<Out> the created bare node
 */
template<typename Out, typename... In>
node_t<Out> extract(record_t<DTypes<In...>>& r_in, std::string address) {
    auto value = std::get<DSType<Out>>(
        std::visit(
            [](auto& node) -> unique_variant_t<var_ret_t<In...>> { return node.value; },
            r_in.map[address]
        )
    );
    return node_t<Out> {
        .value=value,
        .obs=false,
        .interp=NodeReplay(),
        .last_interp=NodeStandard()
    };
}

bool in(std::string a, std::vector<std::string> vec) {
    for (auto& v : vec) {
        if (a == v) return true;
    }
    return false;
}

/**
 * @brief Transfers a newly created node into a record of a compatible type.
 * 
 * @tparam D the underlying distribution type of the created node
 * @tparam Out the parameterization of the record to which the transfer occurs
 * @param r_out the record to which the transfer occurs
 * @param address the address to which to transfer the node
 * @param n the node to be transferred
 * @return record_t<DTypes<Out...>>& a reference to the record to which the transfer occurred
 */
template<typename D, typename... Out>
void transfer(record_t<DTypes<Out...>>& r_out, std::string address, node_t<D>&& n) {
    r_out.map.insert_or_assign(address, n);
    if (!in(address, r_out.addresses)) r_out.addresses.push_back(address);
}

/**
 * @brief Propose a value from distribution of type P into a model node of type D.
 * 
 * This function is intended for use in proposal subclasses only. 
 * 
 * @tparam D the type of the distribution node in which to store the value
 * @tparam P the type from which to sample and score the value
 * @tparam RNG the type of PRNG used
 * @tparam Ts the parameterization of the record
 * @param r the record in which to store a node_t<D>
 * @param address the address at which to store the node
 * @param dist the distribution from which to sample and score the value; the proposal distribution
 * @param rng the PRNG used
 * @return auto the sampled value
 */
template<typename D, typename P, typename RNG, typename... Ts>
auto propose(record_t<DTypes<Ts...>>& r, std::string address, P&& dist, RNG& rng) {
    auto value = dist.sample(rng);
    auto logprob = dist.logprob(value);
    transfer(
        r,
        address,
        (node_t<D>) {
            .value=value,
            .logprob=logprob,
            .obs=false,
            .interp=NodeReplay(),
            .last_interp=NodeStandard()
        }
    );
    return value;
}

#endif  // PROPOSAL_H
